from datetime import timedelta

from pydantic import BaseModel, PositiveFloat, PositiveInt, field_validator, model_validator
from typing_extensions import Self


class Batching(BaseModel):
    """
    Specifies the batching configuration for splitting a large number of attacks into smaller chunks, distributed over
    time.

    Batching provides a way of distributing the load over time with the goal of mitigating CPU, memory, and network
    usage spikes. Setting up batching allows you to divide the list of targets into smaller, equally-sized, and more
    manageable batches.

    Examples:
        In case of **28** targets, the sizes of batches will be: **6, 6, 6, 6, 4**.

        .. code-block:: python

            Batching(count=5, interval=2)

        In case of **28** targets, the sizes of batches will be: **5, 5, 5, 5, 5, 3**.

        .. code-block:: python

            Batching(size=5, interval=2)

    :param size: Size of each batch.
    :type size: int | None
    :param count: Total number of equal-sized batches.
    :type count: int | None
    :param interval: Interval in seconds between processing two consecutive batches.
    :type interval: int | float | timedelta
    """

    size: PositiveInt | None = None
    count: PositiveInt | None = None
    interval: PositiveInt | PositiveFloat | timedelta = timedelta(seconds=1)

    @field_validator("interval", mode="before")
    @classmethod
    def convert_interval(cls, v):
        if isinstance(v, (int, float)):
            return timedelta(seconds=v)
        return v

    @model_validator(mode="after")
    def check_either_size_or_count_set(self) -> Self:
        if self.size is None and self.count is None:
            raise ValueError("Either 'size' or 'count' must be set.")
        if self.size and self.count:
            raise ValueError("Only one of 'size' or 'count' can be set.")
        return self



from enum import Enum


class TargetingStrategy(Enum):
    """
    Targeting strategy that can be used as an alternative to specifying a collection of targets in the exploit
    configuration.

    :cvar AUTO:
        Selects the available targets based on flag IDs provided by the game server, or all targets provided in the
        server configuration if the function does not take `flag_ids` argument.
    :cvar NOP_TEAM:
        Selects the hosts of the NOP team.
    :cvar OWN_TEAM:
        Selects the hosts of your own team.
    """

    AUTO = "auto"
    NOP_TEAM = "nop_team"
    OWN_TEAM = "own_team"


class FlagIdScope(Enum):
    """
    Enumeration representing the scope of ticks for which flag IDs are provided to the exploit function.

    :cvar SINGLE_TICK:
        `flag_ids` object will represent flag IDs relevant to a single service, target, and tick.
        When using `SINGLE_TICK`, each flag ID that successfully returns a flag will be tracked, allowing Avala client
        to skip the attacks that are using the same flag ID (based on exploit alias, target host and flag id value).
        This is the recommended and optimized approach.
    :cvar LAST_N_TICKS:
        `flag_ids` object will contain a list of flag IDs relevant to a single service, target, and the last N ticks.
        In most cases, not necessary and is inefficient due to performing redundant attacks.
    """

    SINGLE_TICK = "single"
    LAST_N_TICKS = "last_n"



import concurrent.futures
import importlib.util
import logging
import re
import time
from datetime import datetime
from multiprocessing import Process
from pathlib import Path
from queue import Empty
from typing import Any, Iterable, Literal

import tzlocal
from apscheduler.schedulers.background import BackgroundScheduler
from httpx import HTTPStatusError, RequestError
from pydantic import AwareDatetime, ValidationError

from .api_client.client import APIClient
from .api_client.schemas import ConnectionConfig, FlagsEnqueueBody, UnscopedFlagIds
from .decorator.schemas import Batching
from .exploit import Exploit
from .logging import colorize, logger, truncate
from .storage.impl import BlobStorage, FlagIdsHashStorage, UnsentFlagStorage


class Avala:
    def __init__(
        self,
        protocol: Literal["http", "https"] = "http",
        host: str = "localhost",
        port: int = 2024,
        name: str = "anon",
        password: str | None = None,
        redis_url: str | None = None,
    ) -> None:
        """
        Initializes the Avala client. The client is responsible for scheduling, setting up, and running exploits.
        Besides launching exploits, client also collects and submits flags to the Avala server, and reduces repeating
        the same attacks by keeping track of used flag IDs.

        :param protocol: Protocol used by the Avala server, defaults to "http"
        :type protocol: Literal[&quot;http&quot;, &quot;https&quot;], optional
        :param host: Host of the Avala server, defaults to "localhost"
        :type host: str, optional
        :param port: Port of the Avala server, defaults to 2024
        :type port: int, optional
        :param name: Name used for worker identification, defaults to "anon"
        :type name: str, optional
        :param password: Password to the Avala server, defaults to None
        :type password: str | None, optional
        :param redis_url: Connection URL of Redis storage used for keeping track of flag IDs and for blob storage in
        exploit functions, defaults to None
        :type redis_url: str | None, optional
        """

        self._connection: ConnectionConfig = ConnectionConfig(
            protocol=protocol,
            host=host,
            port=port,
            username=name,
            password=password,
        )
        self._worker_name: str = name
        self._client: APIClient
        self._scheduler: BackgroundScheduler
        self._blob_storage: BlobStorage | None = BlobStorage(redis_url, "avala_blobs") if redis_url else None
        self._flag_ids_hash_storage: FlagIdsHashStorage | None = (
            FlagIdsHashStorage(redis_url, "avala_flag_hashes") if redis_url else None
        )
        self._unsent_flag_storage: UnsentFlagStorage | None = (
            UnsentFlagStorage(redis_url, "avala_unsent_flags") if redis_url else None
        )

        self._exploit_directories: set[Path] = set()

        self.suppress_logs = False

    def connect(self) -> bool:
        """
        Connects to the Avala server without running it. This method must be called before using any method
        that uses the API client, other than run() as it calls this method internally.
        """
        client = APIClient.connect(self._connection, quiet=self.suppress_logs)
        if client is None:
            return False

        self._client = client
        return True

    def run(self) -> None:
        """
        Runs the Avala client in production mode. This mode operates on a schedule, scanning registered directories for
        functions decorated with the `@exploit` decorator and scheduling attacks based on their configuration,
        including delay and batching settings. Exploits marked as `draft=True` in the decorator are ignored.

        Flags obtained from the executed exploits are submitted to the Avala server. The client keeps track of the
        successful attacks by storing hashes of used flag IDs to skip running the same attacks multiple times. This mode
        runs indefinitely until interrupted.
        """
        self._show_banner()

        self._scheduler = BackgroundScheduler()
        self.connect()

        if self._client is None:
            return

        self._validate_directories()

        logging.getLogger("apscheduler.executors.default").setLevel(logging.CRITICAL)

        if self._unsent_flag_storage is not None:
            self._scheduler.add_job(
                func=self._enqueue_pending_flags,
                trigger="interval",
                seconds=15,
                id="enqueue_pending_flags",
                next_run_time=datetime.now(),
            )

        try:
            self._scheduler.start()
            while True:
                # WARNING: We intentionally avoid scheduling `_schedule_exploits()` with APScheduler.
                # The `pwntools` library used in attack scripts is not compatible with APScheduler's
                # threading model. To work around this, the method is called in a blocking loop,
                # and we use `time.sleep()` to pause until the next tick.
                seconds_before_next_tick = (
                    self._get_next_tick_start() - datetime.now(tzlocal.get_localzone())
                ).total_seconds()
                time.sleep(seconds_before_next_tick)
                self._schedule_exploits()
        except (KeyboardInterrupt, SystemExit):
            print()  # Add a newline after the ^C
            self._scheduler.shutdown()
        finally:
            logger.info("🙌 Thanks for using Avala!")

    def run_exploit(self, exploit_alias: str) -> None:
        """
        Launches a specific exploit based on its alias immediatelly. The purpose of this mode is to run a specific
        exploit independently of the tick schedule in order to allow rapid exploit development. Flags obtained from the
        attack are submitted to the Avala server. The client keeps track of used flag IDs to avoid running the same
        attacks multiple times, unless the exploit's `draft` option is set to True.

        :param exploit_alias: Alias of the exploit to be launched.
        :type exploit_alias: str
        """
        self._check_connection()

        flag_ids = self._fetch_or_load_flag_ids()

        exploit = next((e for e in self._reload_exploits() if e.alias == exploit_alias), None)
        if not exploit:
            logger.error(
                "❌ Exploit with alias <b>{alias}</> not found.",
                alias=colorize(exploit_alias),
            )
            return

        if exploit.takes_flag_ids and flag_ids is None:
            logger.warning(
                "⚠️  Skipping <b>{alias}</> as it requires flag IDs, but they are not available.",
                alias=colorize(exploit.alias),
            )
            return

        if not exploit.setup(
            game=self._client.game,
            flag_ids=flag_ids,
            blob_storage=self._blob_storage,
            flag_ids_hash_storage=self._flag_ids_hash_storage,
            batching=Batching(count=1),
        ):
            return

        self._launch_exploit(exploit)

    def register_directory(self, dir_path: str) -> None:
        """
        Registers a directory containing exploits. The directory path must be either absolute, or relative to the
        module that creates the Avala instance.

        :param dir_path: Path to the directory containing exploits.
        :type dir_path: str
        """
        path = Path(dir_path)
        if path not in self._exploit_directories:
            self._exploit_directories.add(path)

    def get_flag_ids(self) -> UnscopedFlagIds:
        """
        Fetches the current available flag ids fetched by the Avala server.

        :return: Unscoped flag ids covering flag IDs from all services, targets and ticks.
        :rtype: UnscopedFlagIds
        """
        self._check_connection()
        return self._client.fetch_flag_ids()

    def get_services(self) -> set[str]:
        """
        Fetches a set of names of the available services in flag ids.

        :return: Set of service names.
        :rtype: set[str]
        """
        self._check_connection()
        return self._client.fetch_flag_ids().get_service_names()

    def submit_flags(
        self,
        flags: Iterable[str],
        host: str = "unknown",
        service_name: str = "unknown",
        exploit_alias: str = "manual submission",
    ) -> None:
        """
        Sends flags to the server for submission.

        :param flags: Flags to enqueue.
        :type flags: Iterable[str]
        :param host: Host of the target/victim team.
        :type host: str
        :param service_name: Name of the attacked service.
        :type service_name: str
        :param exploit_alias: Alias of the exploit that retrieved the flags.
        :type exploit_alias: str
        """
        self._check_connection()
        self._client.enqueue(flags, host, self._worker_name, service_name, exploit_alias)

    def match_flags(self, output: Any) -> set[str]:
        """
        Matches flags in the attack's result using the flag format defined in the server settings.

        :param output: Any object that may contain flags when converted to a string. This should be the return value of
        an exploit function.
        :type output: Any
        :return: Set of flags extracted from the output.
        :rtype: set[str]
        """
        self._check_connection()
        return set(re.findall(self._client.game.flag_format, str(output)))

    def list_exploits(self) -> list[tuple[str, bool]]:
        """
        Lists aliases and their draft flags of all found exploits.

        :return: List of tuples containing exploit aliases and their draft flags.
        :rtype: list[tuple[str, bool]]
        """
        return [(exploit.alias, exploit.is_draft) for exploit in self._reload_exploits()]

    def update_directory_paths(self, module_path: str) -> None:
        """
        Updates the exploit directory paths to be relative to the given module path. This is useful when running the
        client from a different working directory when using the CLI. Updates _exploit_directories in place.
        """
        abs_module_path = Path(module_path).parent.resolve()
        updated_directories = set()
        for path in self._exploit_directories:
            if not path.is_absolute():
                path = (abs_module_path / path).resolve()
            updated_directories.add(path)
        self._exploit_directories = updated_directories

    def _validate_directories(self) -> None:
        """
        Validates and filters out invalid (non-existent) registered exploit directories. Updates _exploit_directories
        in place.
        """
        valid_directories = set()
        for path in self._exploit_directories:
            if not path.exists() or not path.is_dir():
                logger.error("❌ Directory not found: {path}", path=path)
            else:
                valid_directories.add(path)

        if valid_directories:
            logger.info(
                "📂 Registered exploit directories: <green>{directories}</>",
                directories=", ".join([d.name for d in valid_directories]),
            )
        else:
            logger.error("❌ No exploit directories found! Register at least one directory to start running exploits.")

        self._exploit_directories = valid_directories

    def _fetch_or_load_flag_ids(self, long_poll: bool = False) -> UnscopedFlagIds | None:
        """
        Attempts to fetch flag IDs from the Avala server. If an error occurs while fetching, it tries to load cached
        flag IDs as a fallback. If the cached flag IDs are also not available, it returns None. All errors are logged.

        :param long_poll: If True, waits for the latest flag IDs from the server by long polling. If False, fetches the
        current available flag IDs from the server, defaults to False.
        :type long_poll: bool, optional
        :return: Unscoped flag ids covering flag IDs from all services, targets and ticks, or None if both fetching and
        loading from cache fails.
        :rtype: UnscopedFlagIds | None
        """
        try:
            return self._client.wait_for_flag_ids() if long_poll else self._client.fetch_flag_ids()
        except (HTTPStatusError, RequestError, ValidationError, Exception) as e:
            logger.error(
                "Failed to fetch flag IDs.\n\n<b>{error}</>\n{error_msg}\n",
                error=type(e).__name__,
                error_msg=e,
            )
            try:
                return self._client.get_cached_flag_ids()
            except (ValidationError, FileNotFoundError) as e:
                logger.error(
                    "Failed to load cached flag IDs.\n\n<b>{error}</>\n{error_msg}\n",
                    error=type(e).__name__,
                    error_msg=e,
                )
                return None

    def _reload_exploits(self) -> list[Exploit]:
        """
        Scans the registered directories for Python scripts containing exploits, executes them to compute
        configurations in `@exploit` decorators, and returns a list of exploits that need to be set up, scheduled,
        and run.

        :return: List of exploits to be set up, scheduled, and run afterwards.
        :rtype: list[Exploit]
        """

        exploits: list[Exploit] = []
        exploit_filepaths = (file for directory in self._exploit_directories for file in directory.glob("*.py"))

        for exploit_filepath in exploit_filepaths:
            try:
                spec = importlib.util.spec_from_file_location(exploit_filepath.stem, exploit_filepath.absolute())
                if spec is None:
                    raise Exception("Failed to load module spec.")

                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)  # type: ignore[union-attr]

                for _, func in module.__dict__.items():
                    if callable(func) and hasattr(func, "exploit") and isinstance(func.exploit, Exploit):
                        exploits.append(func.exploit)
            except Exception as e:
                logger.error(
                    "Failed to load exploit from {file}: {error}",
                    file=exploit_filepath,
                    error=e,
                )

        if exploits:
            if not self.suppress_logs:
                logger.info(
                    "📥 Loaded <b>{count}</> exploits: {exploits}",
                    count=len(exploits),
                    exploits=", ".join(
                        (f"{colorize(exploit.alias)}*" if exploit.is_draft else colorize(exploit.alias))
                        for exploit in exploits
                    ),
                )
        else:
            logger.warning("⚠️  No exploits loaded.")

        return exploits

    def _schedule_exploits(self) -> None:
        """
        Handles reloading exploits, setting up, and scheduling them for execution. As a part of setting up exploits,
        it fetches flag IDs from the server by long polling, and passes them to the exploits that require them.

        Should be run as a scheduled job with the interval set to the tick duration.
        """
        executor = concurrent.futures.ThreadPoolExecutor()
        flag_ids_future = executor.submit(self._fetch_or_load_flag_ids, long_poll=True)

        now = datetime.now()

        # Sort exploits so that exploits that don't take flag IDs are scheduled first, followed by exploits that do.
        # This way, fetching flag ids (resolving flag_ids_future) won't block scheduling of exploits that don't need
        # flag ids.

        exploits = [e for e in self._reload_exploits() if not e.is_draft]
        ordered_exploits = [e for e in exploits if not e.takes_flag_ids] + [e for e in exploits if e.takes_flag_ids]

        for exploit in ordered_exploits:
            flag_ids = flag_ids_future.result() if exploit.takes_flag_ids else None
            if exploit.takes_flag_ids and flag_ids is None:
                logger.warning(
                    "⚠️  Skipping <b>{alias}</> as it requires flag IDs, but they are not available.",
                    alias=colorize(exploit.alias),
                )
                continue

            if not exploit.setup(
                game=self._client.game,
                flag_ids=flag_ids,
                blob_storage=self._blob_storage,
                flag_ids_hash_storage=self._flag_ids_hash_storage,
            ):
                continue
            for batch_idx in range(exploit.get_batch_count()):
                run_time = now + exploit.delay + exploit.get_batch_interval() * batch_idx
                self._scheduler.add_job(
                    func=self._launch_exploit,
                    args=(exploit, batch_idx),
                    trigger="date",
                    run_date=run_time,
                    misfire_grace_time=None,
                )
                logger.info(
                    "⏰ Exploit <b>{alias}</> (batch {batch_idx}/{total_batches}) will run at <b>{time}</>.",
                    alias=colorize(exploit.alias),
                    batch_idx=batch_idx + 1,
                    total_batches=exploit.get_batch_count(),
                    time=run_time.strftime("%H:%M:%S"),
                )

        executor.shutdown(wait=True)

    def _launch_exploit(self, exploit: Exploit, batch_idx: int = 0) -> None:
        """
        Launches the exploit in a separate process, collects flags from the attacks as soon as they complete and sends
        them to the Avala server for submission. Also handles timeouts and termination of the exploit process. Each
        exploit gets its own process, and each attack of the exploit gets its own thread.

        :param exploit: Exploit to launch.
        :type exploit: Exploit
        :param batch_idx: Index of the batch of hosts to attack, defaults to 0
        :type batch_idx: int, optional
        """
        logger.info(
            "🚀 Launching exploit <b>{alias}</> (batch {batch_idx}/{total_batches})...",
            alias=colorize(exploit.alias),
            batch_idx=batch_idx + 1,
            total_batches=exploit.get_batch_count(),
        )

        runner = Process(target=exploit.run_attacks, args=(batch_idx,))
        runner.start()

        while runner.is_alive():
            self._collect_and_enqueue(exploit)

        runner.join(exploit.timeout.total_seconds())
        if runner.is_alive():
            logger.info(
                "⌛ Terminating process for <b>{alias}</> (batch {batch_idx}/{total_batches}) due to timeout.",
                alias=colorize(exploit.alias),
                batch_idx=batch_idx + 1,
                total_batches=exploit.get_batch_count(),
            )
            runner.terminate()
            runner.join()
        else:
            logger.info(
                "✅ Finished running <b>{alias}</> (batch {batch_idx}/{total_batches}) in time.",
                alias=colorize(exploit.alias),
                batch_idx=batch_idx + 1,
                total_batches=exploit.get_batch_count(),
            )

        while self._collect_and_enqueue(exploit, drain=True):
            pass

    def _collect_and_enqueue(self, exploit: Exploit, drain: bool = False) -> bool:
        """
        Collects flags from the attacks as soon as they complete and sends them to the Avala server for submission.

        :param exploit: Exploit to collect flags from.
        :type exploit: Exploit
        :param drain: If True, collects all the remaining flags in a non-blocking way, used when all the attacks are
        done. Defaults to False.
        :type drain: bool, optional
        :return: True if there are more flags to collect, False otherwise.
        :rtype: bool
        """
        try:
            host, result = exploit.results.get(
                timeout=None if drain else 0.1,
                block=False if drain else True,
            )
        except Empty:
            return False

        flags = self.match_flags(result)
        if not flags:
            logger.warning(
                "⚠️  No flags retrieved from attacking <b>{host}</> via <b>{alias}</>.",
                host=colorize(host),
                alias=colorize(exploit.alias),
            )
            return True

        try:
            self._client.enqueue(flags, host, self._worker_name, exploit.service, exploit.alias)
        except Exception:
            logger.error(
                "🚨 Failed to submit flags from attacking <b>{host}</> via <b>{alias}</>. <d>{flags}</>",
                host=colorize(host),
                alias=colorize(exploit.alias),
                flags=truncate(", ".join(flags)),
            )
            if self._unsent_flag_storage:
                self._unsent_flag_storage.add(
                    FlagsEnqueueBody(
                        values=flags,
                        host=host,
                        worker=self._worker_name,
                        service=exploit.service,
                        exploit=exploit.alias,
                    )
                )
        finally:
            return True

    def _enqueue_pending_flags(self) -> None:
        """
        Job that periodically checks the connection with the server and tries to enqueue the pending flags collected
        during the server downtime.
        """
        if self._unsent_flag_storage is None:  # check for typing
            return

        if self._unsent_flag_storage.size():
            logger.warning(
                "🔄 <b>{count}</> flags are waiting to be submitted! Checking connection with the server...",
                count=self._unsent_flag_storage.size(),
            )
        else:
            logger.info("👌 No pending flags to submit.")
            return

        try:
            self._client.heartbeat()
        except Exception:
            logger.error(
                "🚨 Cannot establish connection with the server. <b>{count}</> flags are waiting to be submitted!",
                count=self._unsent_flag_storage.size(),
            )
            return
        else:
            logger.info(
                "🔌 Server is back online! Submitting pending <b>{count}</> flags...",
                count=self._unsent_flag_storage.size(),
            )

        flag = self._unsent_flag_storage.pop()
        while flag:
            try:
                self._client.enqueue(
                    flag.values,
                    flag.host,
                    self._worker_name,
                    flag.service or "unknown",
                    flag.exploit or "unknown",
                )
            except Exception:
                self._unsent_flag_storage.add(flag)
                logger.error(
                    "🚨 Cannot establish connection with the server. <b>{count}</> flags are waiting to be submitted!",
                    count=self._unsent_flag_storage.size(),
                )
                break
            else:
                flag = self._unsent_flag_storage.pop()

    def _get_next_tick_start(self) -> AwareDatetime:
        """
        Calculate the start time of the next tick.

        :return: Start time of the next tick.
        :rtype: AwareDatetime
        """
        first_tick_start = self._client.schedule.first_tick_start
        tick_duration = self._client.schedule.tick_duration
        now: AwareDatetime = datetime.now(tzlocal.get_localzone())

        game_has_started = now > first_tick_start

        if not game_has_started:
            return first_tick_start

        return now + tick_duration - (now - first_tick_start) % tick_duration

    def _check_connection(self) -> None:
        """
        Raises a runtime error if the APIClient is not initialized.

        :raises RuntimeError: Avala instance has no connection established.
        """
        if not hasattr(self, "_client"):
            raise RuntimeError(
                "Avala instance has no connection established. Call .connect() method before running this function."
            )

    def _show_banner(self) -> None:
        """
        Shows banner in the terminal. ✨
        """

        print(
            """\033[34;1m
      db
     ;MM:
    ,V^MM. 7MM""Yq.  ,6"Yb.  `7M""MMF',6"Yb.
   ,M  `MM `MM   j8 8)   MM    M  MM 8)   MM
   AbmmmqMA MM""Yq.  ,pm9MM   ,P  MM  ,pm9MM
  A'     VML`M   j8 8M   MM . d'  MM 8M   MM
.AMA.   .AMMA.mmm9' `Moo9^Yo8M' .JMML`Moo9^Yo.
\033[0m"""
        )



from .decorator.decorator import exploit
from .decorator.enums import FlagIdScope, TargetingStrategy
from .decorator.schemas import Batching
from .main import Avala
from .storage.impl import BlobStorage as Store

__all__ = ["Avala", "exploit", "TargetingStrategy", "FlagIdScope", "Batching", "Store"]



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from pydantic import RedisDsn
from redis import Redis

T = TypeVar("T")


class BaseRedisStorage(ABC, Generic[T]):
    """
    Abstract base class for Redis-based stores.
    """

    def __init__(self, redis_url: str = "redis://localhost:6379", name: str = "storage") -> None:
        """
        Initializes the storage with connection to a Redis instance.

        :param redis_url: Redis connection URL, defaults to "redis://localhost:6379"
        :type redis_url: str, optional
        :param name: Name of the Redis storage, defaults to "storage"
        :type name: str, optional
        """
        redis_dsn = RedisDsn(redis_url)

        self._redis = Redis(
            host=redis_dsn.host,
            port=redis_dsn.port,
            password=redis_dsn.password,
            decode_responses=False,
        )
        self._name = name

    @abstractmethod
    def _encode(self, value: T) -> bytes:
        """
        Encodes a value for storage in Redis.

        :param value: The value to encode.
        :type value: T
        :return: The encoded value.
        :rtype: bytes
        """
        pass

    @abstractmethod
    def _decode(self, value: bytes) -> T:
        """
        Decodes a value retrieved from Redis.

        :param value: The value to decode.
        :type value: bytes
        :return: The decoded value.
        :rtype: T
        """
        pass


class HashRedisStorage(BaseRedisStorage[T]):
    """
    Key-value store implemented as a Redis hash.
    """

    def put(self, key: str, value: T, overwrite: bool = True) -> None:
        """
        Stores a key-value pair in Redis hash.

        :param key: The key under which the value will be stored.
        :type key: str
        :param value: The value to store.
        :type value: T
        :param overwrite: Whether to overwrite the existing value. Defaults to True.
        :type overwrite: bool, optional
        """
        if value is None:
            raise ValueError("Cannot store None value.")

        if not overwrite and self._redis.hget(self._name, key) is not None:
            raise KeyError(f"Key '{key}' already exists and overwrite is set to False.")

        encoded_value = self._encode(value)
        self._redis.hset(self._name, key, encoded_value)  # type: ignore

    def get(self, key: str) -> T:
        """
        Retrieves a value associated with the given key from Redis hash.

        :param key: The key to retrieve.
        :type key: str
        :raises KeyError: If the key does not exist.
        :return: The value associated with the key.
        :rtype: T
        """
        value: bytes | None = self._redis.hget(self._name, key)  # type: ignore
        if value is None:
            raise KeyError(f"Key '{key}' not found.")
        return self._decode(value)

    def delete(self, key: str) -> bool:
        """
        Deletes the value associated with the given key from Redis hash.

        :param key: The key to delete.
        :type key: str
        :return: True if the value was deleted, False otherwise.
        :rtype: bool
        """
        if self._redis.hdel(self._name, key):
            return True
        raise KeyError(f"Key '{key}' not found.")

    def contains(self, key: str) -> bool:
        """
        Checks if the given key exists in the Redis hash.

        :param key: The key to check.
        :type key: str
        :return: True if the key exists, False otherwise.
        :rtype: bool
        """
        return self._redis.hexists(self._name, key)  # type: ignore

    def size(self) -> int:
        """
        Returns the size of the Redis hash.

        :return: Size of the hash.
        :rtype: int
        """
        return self._redis.hlen(self._name)  # type: ignore[return-value]

    def __getitem__(self, key: str) -> T:
        return self.get(key)

    def __setitem__(self, key: str, value: T) -> None:
        self.put(key, value)

    def __delitem__(self, key: str) -> None:
        self.delete(key)

    def __contains__(self, key: str) -> bool:
        return self.contains(key)


class SetRedisStorage(BaseRedisStorage[T]):
    """
    Storage for managing a Redis set.
    """

    def add(self, *values: T) -> None:
        """
        Stores one or more values in a Redis set.

        :param values: Values to store.
        :type values: T
        """
        encoded_values = [self._encode(value) for value in values]
        self._redis.sadd(self._name, *encoded_values)

    def remove(self, *values: T) -> None:
        """
        Removes one or more values from a Redis set.

        :param values: Values to remove.
        :type values: T
        """
        encoded_values = [self._encode(value) for value in values]
        self._redis.srem(self._name, *encoded_values)

    def pop(self) -> T | None:
        """
        Retrieves and removes a random value from the Redis set.

        :return: Popped value.
        :rtype: T
        """
        raw_value: bytes | None = self._redis.spop(self._name)  # type: ignore
        return self._decode(raw_value) if raw_value is not None else None

    def contains(self, value: T) -> bool:
        """
        Checks if a value exists in the Redis set.

        :param value: The value to check.
        :type value: T
        :return: True if the value exists, False otherwise.
        :rtype: bool
        """
        encoded_value = self._encode(value)
        return self._redis.sismember(self._name, encoded_value) == 1  # type: ignore

    def size(self) -> int:
        """
        Returns the size of the Redis set.

        :return: Size of the set.
        :rtype: int
        """
        return self._redis.scard(self._name)  # type: ignore[return-value]

    def __contains__(self, value: T) -> bool:
        return self.contains(value)



from datetime import timedelta
from functools import wraps
from typing import Iterable

from ..exploit import Exploit
from .enums import FlagIdScope, TargetingStrategy
from .schemas import Batching


def exploit(
    service: str,
    draft: bool = False,
    alias: str | None = None,
    targets: Iterable[str] | TargetingStrategy = TargetingStrategy.AUTO,
    flag_id_scope: FlagIdScope = FlagIdScope.SINGLE_TICK,
    skip: Iterable[str] | None = None,
    include: Iterable[str] | None = None,
    delay: int | float | timedelta = timedelta(seconds=0),
    batching: Batching = Batching(count=1),
    timeout: int | float | timedelta = timedelta(seconds=15),
):
    """
    Decorator for defining and configuring an exploit.

    :param service: Name of the service attacked by the exploit. To see the names of available services, run `avl services`.
    :type service: str
    :param draft: Exclude the exploit when running Avala in production mode. Useful for testing and debugging exploits when running manually. Defaults to False.
    :type draft: bool
    :param alias: Alias used for exploit identification, logging and as a key for tracking repeated flag IDs.
    :type alias: str | None
    :param targets: IP addresses or hostnames of the targeted teams, or a targeting strategy. Defaults to `TargetingStrategy.AUTO`.
    :type targets: Iterable[str] | TargetingStrategy
    :param flag_id_scope: Tick scope of the flag IDs provided to the exploit function, defaults to
    FlagIdScope.SINGLE_TICK.
    :type flag_id_scope: FlagIdScope
    :param skip: IP addresses or hostnames to skip when attacking. Hosts of the NOP team and own team are skipped by
    default.
    :type skip: Iterable[str] | None
    :param include: Additional IP addresses or hostnames to include when attacking. Can be used to include hosts that
    are skipped by default (NOP team and own team).
    :type include: Iterable[str] | None
    :param delay: Delay in seconds to wait before starting the first attack, defaults to 0. Useful when running
    multiple exploits and need a way to prevent them from running at the same time, which could lead to excessive CPU, memory or network usage.
    :type delay: int | float | timedelta, optional
    :param batching: Batching configuration, defaults to `Batching(count=1)` meaning no batching. Provides a way of distributing the load over time with the goal of mitigating CPU, memory and network usage spikes.
    :type batching: Batching | None, optional
    :param timeout: Timeout in seconds after which the exploit will be terminated if it hangs or takes too long to complete, defaults to 15.
    :type timeout: int, optional
    """  # noqa: E501

    def exploit_decorator(func):
        """
        Creates a runnable `Exploit` object based on the user-defined configuration and associates it with the decorated
        function. Makes the decorated function discoverable by Avala and allows it to setup and run the exploit
        afterwards.
        """

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        wrapper.exploit = Exploit(
            service=service,
            is_draft=draft,
            alias=alias or f"{func.__module__}.{func.__name__}",
            targets_skip=set(skip) if skip else set(),
            targets_include=set(include) if include else set(),
            targets_explicit=set(targets) if isinstance(targets, Iterable) else set(),
            targets_strategy=targets if isinstance(targets, TargetingStrategy) else None,
            flag_id_scope=flag_id_scope,
            delay=timedelta(seconds=delay) if isinstance(delay, (int, float)) else delay,
            batching=batching,
            timeout=timedelta(seconds=timeout) if isinstance(timeout, (int, float)) else timeout,
            func=func,
        )

        return wrapper

    return exploit_decorator



import inspect
import threading
from datetime import timedelta
from multiprocessing import Queue
from typing import Any, Callable

from .api_client.schemas import (
    GameConfig,
    ServiceScopedFlagIds,
    TickScopedFlagIds,
    UnscopedFlagIds,
)
from .decorator.enums import FlagIdScope, TargetingStrategy
from .decorator.schemas import Batching
from .logging import colorize, logger, suggest_closest_match
from .storage.impl import BlobStorage, FlagIdsHashStorage


class Exploit:
    def __init__(
        self,
        service: str,
        is_draft: bool,
        alias: str,
        targets_skip: set[str],
        targets_include: set[str],
        targets_explicit: set[str],
        targets_strategy: TargetingStrategy | None,
        flag_id_scope: FlagIdScope,
        delay: timedelta,
        batching: Batching,
        timeout: timedelta,
        func: Callable,
    ):
        """
        Runnable configured Exploit. Contains processed user-defined configuration and a reference to the exploit
        function.
        """
        self.service = service
        self.is_draft = is_draft
        self.alias = alias
        self.targets_skip = targets_skip
        self.targets_include = targets_include
        self.targets_explicit = targets_explicit
        self.targets_strategy = targets_strategy
        self.flag_id_scope = flag_id_scope
        self.delay = delay
        self.batching = batching
        self.timeout = timeout
        self.func = func

        # Results returned from running the attacks
        self.results: Queue[tuple[str, Any]] = Queue()

        # Helper attributes
        func_params = inspect.signature(func).parameters
        self.takes_flag_ids = "flag_ids" in func_params
        self.takes_storage = "store" in func_params

        # Internal attributes
        self._batched_target_hosts: list[list[str]]
        self._service_flag_ids: ServiceScopedFlagIds | None = None
        self._blob_storage: BlobStorage | None = None
        self._flag_ids_hash_storage: FlagIdsHashStorage | None = None

    def setup(  # noqa: C901
        self,
        *,
        game: GameConfig,
        flag_ids: UnscopedFlagIds | None = None,
        blob_storage: BlobStorage | None = None,
        flag_ids_hash_storage: FlagIdsHashStorage | None = None,
        **overrides,
    ) -> bool:
        """
        Sets up the exploit by validating the passed arguments and computing the target host lists. Returns `True` if
        the setup was successful, otherwise `False`.
        """
        # Override any attributes (commonly used for reseting batching or delay in watch mode).
        for key, value in overrides.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise AttributeError(f"Attempted to override a non-existent attribute '{key}'.")

        # Compute list of the target hosts
        if self.takes_flag_ids:
            if flag_ids is None:
                logger.error(
                    "🚫 Failed to setup {alias} due to missing flag ids.",
                    alias=colorize(self.alias, reset="red"),
                )
                return False
            elif self.service not in flag_ids.get_service_names():
                logger.error(
                    "🚫 Failed to setup {alias} due to invalid service '<b>{service}</>'.",
                    alias=colorize(self.alias, reset="red"),
                    service=self.service,
                )

                suggestion = suggest_closest_match(self.service, flag_ids.get_service_names())
                if suggestion:
                    logger.info(
                        "❓ Did you mean <b>{suggestion}</> instead of '{original}'?",
                        suggestion=suggestion,
                        original=self.service,
                    )
                return False
            else:
                self._service_flag_ids = flag_ids / self.service
                auto_targets = self._service_flag_ids.get_target_hosts()
        else:
            auto_targets = game.opp_team_hosts

        if self.targets_strategy == TargetingStrategy.AUTO:
            target_hosts = (
                auto_targets - (game.own_team_hosts | game.nop_team_hosts | self.targets_skip) | self.targets_include
            )
        elif self.targets_strategy == TargetingStrategy.OWN_TEAM:
            target_hosts = game.own_team_hosts - self.targets_skip | self.targets_include
        elif self.targets_strategy == TargetingStrategy.NOP_TEAM:
            target_hosts = game.nop_team_hosts - self.targets_skip | self.targets_include
        else:  # Manual targets
            target_hosts = (
                self.targets_explicit - (game.own_team_hosts | game.nop_team_hosts | self.targets_skip)
                | self.targets_include
            )
            if self._service_flag_ids is not None:
                target_hosts &= self._service_flag_ids.get_target_hosts()

        # Organize target hosts in batches
        if self.batching.size:
            self._batched_target_hosts = self._batch_target_hosts_by_size(target_hosts)
        elif self.batching.count:
            self._batched_target_hosts = self._batch_target_hosts_by_count(target_hosts)

        # Set up flag IDs hash storage
        self._flag_ids_hash_storage = flag_ids_hash_storage

        # Set up blob storage
        if self.takes_storage:
            if blob_storage is None:
                logger.error(
                    "🚫 Failed to setup {alias} due to missing Redis storage.",
                    alias=colorize(self.alias, reset="red"),
                )
                return False
            else:
                self._blob_storage = blob_storage

        return True

    def run_attacks(self, batch_idx: int = 0) -> None:  # noqa: C901
        """
        Executes attacks against the hosts in the specified batch, with each attack running in its own separate thread.
        Running attacks also includes putting the results into the results queue.

        :param batch_idx: Index of the batch of target hosts, defaults to 0. Use 0 if batching is disabled (there is
        only one batch).
        :type batch_idx: int, optional
        """
        if not hasattr(self, "_batched_target_hosts"):
            logger.error(
                "🚫 Skipped {alias} due to setup failure (_batched_target_hosts missing).",
                alias=colorize(self.alias, reset="red"),
            )
            return

        target_hosts = self._batched_target_hosts[batch_idx]
        if not target_hosts:
            logger.warning(
                "⏩ Skipping {alias} due to no targets specified.",
                alias=colorize(self.alias),
            )
            return

        attack_threads = []
        skipped_attacks = []

        for host in target_hosts:
            if self.takes_flag_ids and self._service_flag_ids is not None:
                if self.flag_id_scope == FlagIdScope.SINGLE_TICK:
                    for single_tick_flag_id in (self._service_flag_ids / host).walk():
                        if (
                            self._flag_ids_hash_storage
                            and not self.is_draft
                            and single_tick_flag_id.compute_hash(self.alias) in self._flag_ids_hash_storage
                        ):
                            skipped_attacks.append(host)
                            continue
                        thread = threading.Thread(
                            target=self._run_attack,
                            args=(host, single_tick_flag_id),
                        )
                        attack_threads.append(thread)
                        thread.start()
                elif self.flag_id_scope == FlagIdScope.LAST_N_TICKS:
                    single_tick_flag_id_values = []
                    for single_tick_flag_id in (self._service_flag_ids / host).walk():
                        if (
                            self._flag_ids_hash_storage
                            and not self.is_draft
                            and single_tick_flag_id.compute_hash(self.alias) in self._flag_ids_hash_storage
                        ):
                            skipped_attacks.append(host)
                        else:
                            single_tick_flag_id_values.append(single_tick_flag_id.value)

                    thread = threading.Thread(
                        target=self._run_attack,
                        args=(host, single_tick_flag_id_values),
                    )
                    attack_threads.append(thread)
                    thread.start()
            else:
                thread = threading.Thread(
                    target=self._run_attack,
                    args=(host, None),
                )
                attack_threads.append(thread)
                thread.start()

        if skipped_attacks:
            logger.info(
                "⏩ Skipped total <b>{count}</> completed attacks via <b>{alias}</> on hosts {hosts}.",  # noqa: E501
                count=len(skipped_attacks),
                alias=colorize(self.alias),
                hosts=", ".join([colorize(host) for host in set(skipped_attacks)]),
            )

        try:
            for thread in attack_threads:
                thread.join()
        except (KeyboardInterrupt, SystemExit):
            logger.info(
                "🛑 Stopping {alias} (batch {batch_idx}/{batch_count}) due to user interruption.",
                alias=colorize(self.alias),
                batch_idx=batch_idx + 1,
                batch_count=len(self._batched_target_hosts),
            )

    def get_batch_count(self) -> int:
        """
        Returns the number of batches that will be run.

        :return: Number of batches.
        :rtype: int
        """
        return len(self._batched_target_hosts)

    def get_batch_interval(self) -> timedelta:
        """
        Returns the interval between batches.

        :return: Interval between batches.
        :rtype: timedelta
        """
        return self.batching.interval  # type: ignore[return-value]

    def _batch_target_hosts_by_size(self, hosts_set: set[str]) -> list[list[str]]:
        """
        Groups target hosts into batches of a specified size.

        :param hosts_set: Set of hosts to be distributed into batches.
        :type hosts_set: set[str]
        :return: List of batches, where each batch is a list of hosts.
        :rtype: list[list[str]]
        """
        if hosts_set is None or self.batching is None or self.batching.size is None:
            return []

        hosts_list = list(hosts_set)

        return [hosts_list[i : i + self.batching.size] for i in range(0, len(hosts_list), self.batching.size)]

    def _batch_target_hosts_by_count(self, hosts_set: set[str]) -> list[list[str]]:
        """
        Distributes target hosts into a specified number of batches as evenly as possible.

        :param hosts_set: Set of hosts to be distributed into batches.
        :type hosts_set: set[str]
        :return: List of batches, where each batch is a list of hosts.
        :rtype: list[list[str]]
        """
        if hosts_set is None or self.batching is None or self.batching.count is None:
            return []

        hosts_list = list(hosts_set)

        size = len(hosts_list) // self.batching.count
        remainder = len(hosts_list) % self.batching.count
        batches = [hosts_list[i * size : (i + 1) * size] for i in range(self.batching.count)]
        for i in range(remainder):
            batches[i].append(hosts_list[self.batching.count * size + i])

        return [batch for batch in batches if batch]

    def _is_first_batch(self, batch_idx: int):
        return batch_idx == 0

    def _is_last_batch(self, batch_idx: int):
        return batch_idx == len(self._batched_target_hosts) - 1

    def _run_attack(self, host: str, flag_ids: TickScopedFlagIds | list | None = None):
        """
        Runs a single attack against the specified target host.

        :param host: Target host to attack.
        :type host: str
        :param flag_ids: TickScopedFlagIds object if , defaults to None
        :type flag_ids: TickScopedFlagIds | list | None, optional
        """
        kwargs: dict[str, Any] = {"target": host}
        if self.takes_storage:
            kwargs["store"] = self._blob_storage
        if self.takes_flag_ids:
            kwargs["flag_ids"] = flag_ids.value if isinstance(flag_ids, TickScopedFlagIds) else flag_ids

        try:
            result = self.func(**kwargs)
        except Exception as e:
            logger.error(
                "An error has occured while attacking <b>{host}</> via <b>{alias}</>: {error}",
                host=colorize(host, reset="red"),
                alias=colorize(self.alias, reset="red"),
                error=e,
            )
        else:
            self.results.put((host, result))

            if isinstance(flag_ids, TickScopedFlagIds) and not self.is_draft and self._flag_ids_hash_storage:
                self._flag_ids_hash_storage.add(flag_ids.compute_hash(self.alias))



import json
from pathlib import Path
from typing import Iterable, Optional

import httpx

from ..logging import colorize, logger, truncate
from .schemas import (
    ConnectionConfig,
    FlagsEnqueueBody,
    FlagsEnqueueResponse,
    GameConfig,
    ScheduleConfig,
    UnscopedFlagIds,
)

DOT_DIR_PATH = Path(".avala")


class APIClient:
    """
    Class for interacting with the Avala server API and keeping configuration for the
    game and scheduling.
    """

    def __init__(
        self,
        connection: ConnectionConfig,
    ) -> None:
        self.connection: ConnectionConfig = connection
        self.client: httpx.Client = self._setup_http_client()
        self.game: GameConfig
        self.schedule: ScheduleConfig
        self.game, self.schedule = self._fetch_settings()
        DOT_DIR_PATH.mkdir(exist_ok=True)

    @classmethod
    def connect(cls, connection: ConnectionConfig, quiet: bool = False) -> Optional["APIClient"]:
        """
        Connect to the Avala server and return an APIClient instance, or return None and
        print an error message if the connection fails.

        :param connection: Connection configuration for the Avala server.
        :type connection: ConnectionConfig
        :param quiet: If True, suppresses the log messages upon a successful connection, defaults to False.
        :type quiet: bool
        :return: An instance of APIClient.
        :rtype: APIClient | None
        """
        try:
            instance = cls(connection)

            if not quiet:
                logger.success(
                    "✅ Connected to Avala server at <b>{protocol}://{host}:{port}</> as <b>{username}</>.",
                    protocol=connection.protocol,
                    host=connection.host,
                    port=connection.port,
                    username=connection.username,
                )

                logger.info(
                    "Flag format: <b>{flag_format}</>",
                    flag_format=instance.game.flag_format,
                )

                logger.info(
                    "Tick duration: <b>{tick_duration}</> seconds",
                    tick_duration=instance.schedule.tick_duration.total_seconds(),
                )

                logger.info(
                    "Time of the first tick: <b>{first_tick_timestamp}</>",
                    first_tick_timestamp=instance.schedule.first_tick_start,
                )

            return instance
        except Exception as e:
            logger.error(
                "Failed to connect to Avala server.\n\n<b>{error}</>\n{error_msg}\n",
                error=type(e).__name__,
                error_msg=e,
            )
            logger.info("❌ Exiting...")
            return None

    def heartbeat(self) -> None:
        """
        Check if the client is still connected to the server.

        :raises RuntimeError: If the connection was never established.
        :raises httpx.HTTPStatusError: If the server is unreachable within 5 seconds or
        responds with an error status code.
        """
        self.client.get("/health", timeout=5).raise_for_status()

    def enqueue(
        self,
        flags: Iterable[str],
        host: str,
        worker_name: str,
        service_name: str,
        exploit_alias: str,
    ) -> None:
        """
        Sends flags to the server for submission.

        :param flags: Flags to enqueue.
        :type flags: Iterable[str]
        :param host: Host of the target/victim team.
        :type host: str
        :param worker_name: Name of the worker that retrieved the flags.
        :type worker_name: str
        :param service_name: Name of the attacked service.
        :type service_name: str
        :param exploit_alias: Alias of the exploit that retrieved the flags.
        :type exploit_alias: str
        :raises httpx.HTTPStatusError: If the server responds with an error status code.
        """
        enqueue_body = FlagsEnqueueBody(
            values=set(flags),
            host=host,
            service=service_name,
            worker=worker_name,
            exploit=exploit_alias,
        )

        response = self.client.post(
            "/flags",
            json=enqueue_body.model_dump(mode="json"),
        )
        response.raise_for_status()

        flag_enqueue_response = FlagsEnqueueResponse(**response.json())

        logger.info(
            "{icon} Enqueued <b>{enqueued}/{total}</> flags from <b>{host}</> via <b>{exploit}</>. <d>{flags}</>",
            icon="🚩" if flag_enqueue_response.enqueued else "❗",
            enqueued=flag_enqueue_response.enqueued,
            total=len(list(flags)),
            host=colorize(host),
            exploit=colorize(exploit_alias),
            flags=truncate(", ".join(flags)),
        )

    def wait_for_flag_ids(self) -> UnscopedFlagIds:
        """
        Waits for the latest flag ids from the server by long polling. Useful for starting
        the attacks using the latest up-to-date flag ids.

        :raises httpx.HTTPStatusError: If the server responds with an error status code.
        :return: Unscoped flag ids covering flag IDs from all services, targets and ticks.
        :rtype: UnscopedFlagIds
        """
        response = self.client.get(
            "/flag-ids",
            params={"wait": True},
            timeout=self.schedule.tick_duration.total_seconds()
            - 1,  # Subtract 1 second to prevent scheduler jobs from overlapping
        )
        response.raise_for_status()

        if response.status_code == 200:
            self._cache_flag_ids(response.json())

        return UnscopedFlagIds(response.json())

    def fetch_flag_ids(self) -> UnscopedFlagIds:
        """
        Fetches the current available flag IDs from the server. Useful for starting the attacks immediately using the
        currently available flag IDs.

        :raises httpx.HTTPStatusError: If the server responds with an error status code.
        :return: Unscoped flag ids covering flag IDs from all services, targets and ticks.
        :rtype: UnscopedFlagIds
        """
        response = self.client.get("/flag-ids")
        response.raise_for_status()

        if response.status_code == 200:
            self._cache_flag_ids(response.json())

        return UnscopedFlagIds(response.json())

    def get_cached_flag_ids(self) -> UnscopedFlagIds:
        """
        Uses the cached flag IDs as a fallback in case of connection loss or server downtime.

        :raises FileNotFoundError: Flag IDs were never fetched.
        :raises RuntimeError: Flag IDs are corrupted or were never fetched.
        :return: Unscoped flag IDs covering flag IDs from all services, targets and
        ticks.
        :rtype: UnscopedFlagIds
        """
        logger.warning("⚠️  Using cached flag IDs.")

        if not (DOT_DIR_PATH / "cached_flag_ids.json").exists():
            raise FileNotFoundError("Flag IDs were never fetched.")

        with open(DOT_DIR_PATH / "cached_flag_ids.json") as file:
            return UnscopedFlagIds(json.load(file))

    def _setup_http_client(self) -> httpx.Client:
        """
        Sets up the HTTP client for interacting with the Avala API server.

        :raises ConnectionError: If the client fails to establish a connection to the server.
        :return: HTTP client configured for interacting with the server.
        :rtype: httpx.Client
        """
        auth = httpx.BasicAuth(self.connection.username, self.connection.password) if self.connection.password else None

        client = httpx.Client(
            auth=auth,
            base_url=f"{self.connection.protocol}://{self.connection.host}:{self.connection.port}",
        )
        client.get("/health", timeout=5).raise_for_status()

        return client

    def _fetch_settings(self) -> tuple[GameConfig, ScheduleConfig]:
        try:
            response = self.client.get("/configure").json()
            game_data = response.get("game", {})
            schedule_data = response.get("schedule", {})
            return (
                GameConfig.model_validate(game_data),
                ScheduleConfig.model_validate(schedule_data),
            )
        except Exception as e:
            logger.error(
                "Failed to fetch and parse configuration.\n\n<b>{error}</>\n{error_msg}\n",
                error=type(e).__name__,
                error_msg=e,
            )
            raise

    def _cache_flag_ids(self, response_json: dict) -> None:
        """
        Caches the fetched flag IDs to a JSON file as a temporary fallback in case of
        connection loss or server downtime.

        :param response_json: Dictionary containing the fetched flag IDs.
        :type response_json: dict
        """
        with open(DOT_DIR_PATH / "cached_flag_ids.json", "w") as file:
            json.dump(response_json, file)



import importlib.util
import io
import json
import os
import sys
from types import ModuleType
from typing import Any

import click

from .logging import colorize
from .main import Avala


def _import_module_from_path(path: str) -> ModuleType:
    """Import a Python module from a filesystem path."""
    path = os.path.abspath(path)
    if os.path.isdir(path):
        init_path = os.path.join(path, "__init__.py")
        if not os.path.exists(init_path):
            raise FileNotFoundError(f"No __init__.py found in directory: {path}")
        module_name = os.path.basename(path)
        spec = importlib.util.spec_from_file_location(module_name, init_path)
    else:
        if not os.path.exists(path):
            raise FileNotFoundError(f"File not found: {path}")
        module_name = os.path.splitext(os.path.basename(path))[0]
        spec = importlib.util.spec_from_file_location(module_name, path)

    if spec is None or spec.loader is None:
        raise ImportError(f"Could not load module from path: {path}")

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)  # type: ignore[assignment]
    return module


def _find_avala_instance(mod: ModuleType) -> Any:
    """Return an object in the module that is an instance of Avala."""
    for _, obj in vars(mod).items():
        if obj is None:
            continue
        if isinstance(obj, Avala):
            return obj
    raise AttributeError(
        "Could not find an instance of 'Avala' in the imported module. "
        "Please ensure your file defines an Avala instance."
    )


def _load_avala(path: str | None = None) -> Any:
    """
    Load an Avala instance from:
    - the given --path (file or package directory), or
    - './app.py' by default.
    """
    target = path or os.path.join(os.getcwd(), "app.py")
    mod = _import_module_from_path(target)
    return _find_avala_instance(mod)


def _stdin_has_data() -> bool:
    return not sys.stdin.isatty()


def _require_avala(ctx: click.Context) -> Any:
    obj: Context | None = getattr(ctx, "obj", None)  # type: ignore[assignment]
    if obj is None or obj.avala is None:
        tried = getattr(obj, "module_path", os.path.join(os.getcwd(), "app.py")) if obj else "app.py"
        message = [
            "No Avala instance available.",
            f"Tried to import: {tried}",
        ]
        if obj and obj.load_error:
            message.append(f"Cause: {obj.load_error}")
        message.append(
            "\nTips:\n"
            "  - Ensure your module defines an Avala instance at module scope.\n"
            "  - Use --path to point to a different file, e.g. --path ./hello.py\n"
            "  - Run 'avl init' to scaffold a client."
        )
        raise click.ClickException("\n".join(message))
    return obj.avala


class Context:
    def __init__(
        self,
        avala: Avala,
        module_path: str,
        load_error: str | None = None,
    ) -> None:
        self.avala = avala
        self.module_path = module_path
        self.load_error = load_error


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "--path",
    "path_",
    type=click.Path(exists=False, dir_okay=True, file_okay=True, readable=True, path_type=str),
    default=None,
    help="Path to a Python file (e.g. ./hello.py) or package that exports an Avala instance. Defaults to ./app.py.",
)
@click.pass_context
def avl(ctx: click.Context, path_: str | None = None) -> None:
    """
    Avala CLI utility.
    Uses an Avala instance from ./app.py. Use --path to point to a different file or package directory.
    """
    module_path = os.path.abspath(path_ or os.path.join(os.getcwd(), "app.py"))
    try:
        avala = _load_avala(path_)
        ctx.obj = Context(avala=avala, module_path=module_path)
    except Exception as e:
        ctx.obj = Context(avala=None, module_path=module_path, load_error=str(e))  # type: ignore[arg-type]
        return


@avl.command()
@click.pass_context
def init(ctx: click.Context) -> None:
    """
    Guided initialization for a new Avala client (creates app.py and exploit directory).
    """
    protocol = click.prompt("protocol", default="http")
    host = click.prompt("host", default="localhost")
    port = click.prompt("port", default=2024, type=int)
    name = click.prompt("name", default="anon")
    password = click.prompt("password", default="", hide_input=True)
    redis_url = click.prompt("redis url", default="")
    exploit_dir = click.prompt("exploit directory", default="sploits")

    os.makedirs(exploit_dir, exist_ok=True)

    args = [
        f'host="{host}"',
        f"port={port}",
        f'name="{name}"',
        f'password="{password}"',
    ]
    if protocol:
        args.append(f'protocol="{protocol}"')
    if redis_url:
        args.append(f'redis_url="{redis_url}"')
    args_str = ",\n    ".join(args)
    app_content = f"""from avala import Avala

avl = Avala(
    {args_str},
)

avl.register_directory("{exploit_dir}")

if __name__ == "__main__":
    avl.run()
"""

    app_path = "app.py"
    if os.path.exists(app_path):
        overwrite = click.confirm(f"⚠️  '{app_path}' already exists. Overwrite?", default=False)
        if not overwrite:
            click.echo("❌ Initialization cancelled.")
            return

    with open(app_path, "w", encoding="utf-8") as f:
        f.write(app_content)

    click.echo("\n🚀 Initialization complete. Run 'avl run' to run Avala.")


@avl.command()
@click.pass_context
def run(ctx: click.Context) -> None:
    """
    Runs Avala client in production mode.
    """
    avala = _require_avala(ctx)
    avala.update_directory_paths(ctx.obj.module_path)
    avala.run()


@avl.command()
@click.pass_context
def services(ctx: click.Context) -> None:
    """
    Displays all service names based on flag IDs.
    """
    avala = _require_avala(ctx)
    avala.suppress_logs = True
    avala.connect()

    for name in sorted(avala.get_services()):
        click.echo(name)


@avl.command()
@click.argument("service", required=False)
@click.argument("target", required=False)
@click.argument("tick_index", required=False, type=int)
@click.pass_context
def flag_ids(ctx: click.Context, service: str | None, target: str | None, tick_index: int | None) -> None:
    """
    Filters and lists flag IDs.
    """
    avala = _require_avala(ctx)
    avala.suppress_logs = True
    avala.connect()

    if service is None:
        click.echo(json.dumps(avala.get_flag_ids().serialize()))
    elif service and target is None:
        click.echo(json.dumps((avala.get_flag_ids() / service).serialize()))
    elif service and target and tick_index is None:
        click.echo(json.dumps((avala.get_flag_ids() / service / target).serialize()))
    elif service and target and tick_index is not None:
        click.echo((avala.get_flag_ids() / service / target / tick_index).value)


@avl.command()
@click.pass_context
def exploits(ctx: click.Context) -> None:
    """
    Displays aliases of all found exploits.
    """
    avala = _require_avala(ctx)
    avala.suppress_logs = True
    avala.update_directory_paths(ctx.obj.module_path)
    avala.connect()

    for exploit_alias, is_draft in avala.list_exploits():
        draft_marker = " (draft)" if is_draft else ""
        click.echo(f"{colorize(exploit_alias)}{draft_marker}")


@avl.command()
@click.argument("alias", required=True)
@click.pass_context
def launch(ctx: click.Context, alias: str) -> None:
    """
    Launches attacks using an exploit with specified alias.
    """
    avala = _require_avala(ctx)
    avala.suppress_logs = True
    avala.update_directory_paths(ctx.obj.module_path)
    avala.connect()

    avala.run_exploit(alias)


@avl.command()
@click.pass_context
def submit(ctx: click.Context) -> None:
    """
    Extracts flags from text (stdin or interactive paste) and sends them for submission.
    """
    avala = _require_avala(ctx)
    avala.suppress_logs = True
    avala.connect()

    if _stdin_has_data():
        content = sys.stdin.read()
    else:
        click.echo("Paste content containing the flags and press Ctrl-D/Ctrl-Z when done:")
        # Read until EOF (Ctrl-D/Ctrl-Z)
        buf = io.StringIO()
        try:
            for line in sys.stdin:
                buf.write(line)
        except KeyboardInterrupt:
            pass
        content = buf.getvalue()

    if not content.strip():
        raise click.UsageError("No input provided. Pipe or paste content containing flags.")

    flags = avala.match_flags(content)
    click.echo(f"Found {len(flags)} unique flag(s):")
    for flag in flags:
        click.echo(flag)

    avala.submit_flags(flags)


def main() -> None:
    avl(prog_name="avl")


if __name__ == "__main__":
    main()



import hashlib
from datetime import timedelta
from typing import Any, Iterator, Literal

from pydantic import AwareDatetime, BaseModel, Field, PositiveInt


class GameConfig(BaseModel):
    flag_format: str
    own_team_hosts: set[str]
    nop_team_hosts: set[str]
    opp_team_hosts: set[str]


class ScheduleConfig(BaseModel):
    first_tick_start: AwareDatetime
    tick_duration: timedelta
    network_open_tick: PositiveInt
    total_ticks: PositiveInt


class ConnectionConfig(BaseModel):
    protocol: Literal["http", "https"] = "http"
    host: str = "localhost"
    port: int = Field(ge=1, le=65535, default=2024)
    username: str = Field(max_length=20, default="anon")
    password: str | None = None


class FlagsEnqueueBody(BaseModel):
    values: set[str]
    host: str
    worker: str
    service: str | None = None
    exploit: str | None = None


class FlagsEnqueueResponse(BaseModel):
    enqueued: int
    discarded: int


class TickScopedFlagIds:
    """
    Flag IDs for a particular service, target and a tick as provided by the game server.
    """

    def __init__(
        self,
        service_name: str,
        target_host: str,
        ticks_ago: int,
        flag_ids: Any,
    ):
        self._validate(service_name, target_host, ticks_ago)
        self.service_name: str = service_name
        self.target_host: str = target_host
        self.ticks_ago: int = ticks_ago
        self.value: Any = flag_ids

    def compute_hash(self, exploit_alias: str) -> str:
        """
        Hashes the specific flag ID in order to track it to ensure that the same attack
        is not executed multiple times.

        :return: Hash computed from the alias, target, and flag IDs.
        :rtype: str
        """
        return hashlib.md5((exploit_alias + self.target_host + str(self.value)).encode()).hexdigest()

    @staticmethod
    def _validate(service_name: Any, target_host: Any, ticks_ago: Any) -> None:
        if not isinstance(service_name, str):
            raise ValueError(
                "Service name must be a string." + f" Got {type(service_name).__name__} ('{service_name}')."
            )
        if not isinstance(target_host, str):
            raise ValueError("Target host must be a string." + f" Got {type(target_host).__name__} ('{target_host}').")
        if not isinstance(ticks_ago, int):
            raise ValueError(
                "Ticks ago value must be an integer." + f" Got {type(ticks_ago).__name__} ('{ticks_ago}')."
            )


class TargetScopedFlagIds:
    """
    Flag IDs for a particular service and target, and last N ticks provided by the game server. Allows traversing the
    flag IDs structure in a intuitive way.

    Example:
        Using `/` operator you can reduce the scope of the flag IDs to a specific tick. The class hierarchy is as
        follows:

            `UnscopedFlagIds` -> `ServiceScopedFlagIds` -> `TargetScopedFlagIds` -> `TickScopedFlagIds` -> `Any`

        The following example demonstrates how to access flag IDs for a specific service, target, and tick:

        .. code-block:: python

            foo_13_flag_ids: TargetScopedFlagIds  # Flag IDs for the service 'foo' and target '10.10.13.37'
            flag_ids: Any = foo_13_flag_ids / 0   # Flag IDs of the latest tick
    """

    def __init__(
        self,
        service_name: str,
        target_host: str,
        ticks_data: list[Any],
    ):
        self._validate(service_name, target_host, ticks_data)
        self.service_name: str = service_name
        self.target_host: str = target_host
        self.ticks: list[TickScopedFlagIds] = [
            TickScopedFlagIds(service_name, target_host, ticks_ago, flag_ids)
            for ticks_ago, flag_ids in enumerate(ticks_data)
        ]

    def serialize(self) -> list[Any]:
        return [tick.value for tick in self.ticks]

    def get_flag_ids_for_tick(self, index: int) -> TickScopedFlagIds:
        """
        Returns the flag ids for a specific tick.

        :param index: Index of the tick.
        :type index: int
        :return: Flag ids for the specified tick.
        :rtype: TickScopedFlagIds
        :raises IndexError: If the tick index is out of range.
        """
        if 0 <= index < len(self.ticks):
            return self.ticks[index]
        else:
            raise IndexError(f"Tick index '{index}' out of range")

    def walk(self) -> Iterator[TickScopedFlagIds]:
        for tick in self.ticks:
            yield tick

    def copy(self) -> "TargetScopedFlagIds":
        """
        Creates a deep copy of the flag IDs.

        :return: A deep copy of the flag IDs.
        :rtype: TargetScopedFlagIds
        """
        return TargetScopedFlagIds(self.service_name, self.target_host, self.serialize())

    @staticmethod
    def _validate(service_name: Any, target_host: Any, ticks_data: Any) -> None:
        if not isinstance(service_name, str):
            raise ValueError(
                "Service name must be a string." + f" Got {type(service_name).__name__} ('{service_name}')."
            )
        if not isinstance(target_host, str):
            raise ValueError("Target host must be a string." + f" Got {type(target_host).__name__} ('{target_host}').")
        if not isinstance(ticks_data, list):
            raise ValueError(
                f"Flag IDs for the service '{service_name}' and target '{target_host}' must be a list."
                + f" Got {type(ticks_data).__name__} ('{ticks_data}')."
            )

    def __truediv__(self, index: int) -> Any:
        return self.get_flag_ids_for_tick(index)

    def __getitem__(self, index: int) -> Any:
        return self.get_flag_ids_for_tick(index)


class ServiceScopedFlagIds:
    """
    Flag IDs for a particular service, its targets, and last N ticks provided by the game server. Allows traversing the
    flag IDs structure in a intuitive way.

    Example:
        Using `/` operator you can reduce the scope of the flag IDs to a specific target or tick. The class
        hierarchy is as follows:

            `UnscopedFlagIds` -> `ServiceScopedFlagIds` -> `TargetScopedFlagIds` -> `TickScopedFlagIds` -> `Any`

        The following example demonstrates how to access flag IDs for a specific service, target, and tick:

        .. code-block:: python

            foo_flag_ids: ServiceScopedFlagIds
            flag_ids: Any = foo_flag_ids / "10.10.13.37" / 0
    """

    def __init__(
        self,
        service_name: str,
        targets_data: dict[str, list[Any]],
    ):
        self._validate(service_name, targets_data)
        self.service_name: str = service_name
        self.targets: set[TargetScopedFlagIds] = {
            TargetScopedFlagIds(service_name, target_host, ticks_data)
            for target_host, ticks_data in targets_data.items()
        }

        self._target_host_map: dict[str, TargetScopedFlagIds] = {target.target_host: target for target in self.targets}

    def serialize(self) -> dict[str, list[Any]]:
        return {target.target_host: target.serialize() for target in self.targets}

    def get_target_hosts(self) -> set[str]:
        """
        Returns a set of all targets for which flag ids are available.

        :return: Set of IP addresses or hostnames of the targets.
        :rtype: set[str]
        """
        return set(self._target_host_map.keys())

    def get_flag_ids_for_target(self, target_host: str) -> TargetScopedFlagIds:
        """
        Returns the flag ids for a specific target.

        :param target_host: IP address or hostname of the target/victim team.
        :type target_host: str
        :return: Flag ids for the specified target.
        :rtype: TargetScopedFlagIds
        :raises KeyError: If the target is not found.
        """
        if target_host in self._target_host_map:
            return self._target_host_map[target_host]
        else:
            raise KeyError(f"Target '{target_host}' not found")

    def walk(self) -> Iterator[TickScopedFlagIds]:
        for target in self.targets:
            yield from target.walk()

    def copy(self) -> "ServiceScopedFlagIds":
        """
        Creates a deep copy of the flag IDs.

        :return: A deep copy of the flag IDs.
        :rtype: ServiceScopedFlagIds
        """
        return ServiceScopedFlagIds(self.service_name, self.serialize())

    @staticmethod
    def _validate(service_name: Any, targets_data: Any) -> None:
        if not isinstance(service_name, str):
            raise ValueError(
                "Service name must be a string." + f"Got {type(service_name).__name__} ('{service_name}')."
            )
        if not isinstance(targets_data, dict):
            raise ValueError(
                f"Flag IDs of the service '{service_name}' must be a dictionary."
                + f" Got {type(targets_data).__name__} ('{targets_data}')."
            )

    def __truediv__(self, target_host: str) -> TargetScopedFlagIds:
        return self.get_flag_ids_for_target(target_host)

    def __getitem__(self, target_host: str) -> TargetScopedFlagIds:
        return self.get_flag_ids_for_target(target_host)


class UnscopedFlagIds:
    """
    Flag IDs for all services, targets, and last N ticks provided by the game server. Allows traversing the flag IDs
    structure in a intuitive way.

    Example:
        Using `/` operator you can reduce the scope of the flag IDs to a specific service, target, or tick. The class
        hierarchy is as follows:

            `UnscopedFlagIds` -> `ServiceScopedFlagIds` -> `TargetScopedFlagIds` -> `TickScopedFlagIds` -> `Any`

        The following example demonstrates how to access flag IDs for a specific service, target, and tick:

        .. code-block:: python

            all_flag_ids: UnscopedFlagIds
            flag_ids: Any = all_flag_ids / "service_foo" / "10.10.13.37" / 0
    """

    def __init__(self, data: dict[str, dict[str, list[Any]]]):
        self._validate(data)
        self.services: set[ServiceScopedFlagIds] = {
            ServiceScopedFlagIds(service_name, targets_data) for service_name, targets_data in data.items()
        }

        self._service_name_map: dict[str, ServiceScopedFlagIds] = {
            service.service_name: service for service in self.services
        }

    def serialize(self) -> dict[str, dict[str, list[Any]]]:
        return {service.service_name: service.serialize() for service in self.services}

    def get_service_names(self) -> set[str]:
        """
        Returns a set of all service names for which flag ids are available.

        :return: Set of service names.
        :rtype: set[str]
        """
        return set(self._service_name_map.keys())

    def get_flag_ids_for_service(self, service_name: str) -> ServiceScopedFlagIds:
        """
        Returns the flag ids for a specific service and all its targets.

        :param service_name: Name of the service.
        :type service_name: str
        :return: Flag ids for the specified service.
        :rtype: ServiceScopedFlagIds
        :raises KeyError: If the service is not found.
        """
        if service_name in self._service_name_map:
            return self._service_name_map[service_name]
        else:
            raise KeyError(f"Service '{service_name}' not found")

    def walk(self) -> Iterator[TickScopedFlagIds]:
        for service in self.services:
            yield from service.walk()

    def copy(self) -> "UnscopedFlagIds":
        """
        Creates a deep copy of the flag IDs.

        :return: A deep copy of the flag IDs.
        :rtype: UnscopedFlagIds
        """
        return UnscopedFlagIds(self.serialize())

    @staticmethod
    def _validate(data: Any) -> None:
        if not isinstance(data, dict):
            raise ValueError("Flag IDs must be a dictionary.")

    def __truediv__(self, service_name: str) -> ServiceScopedFlagIds:
        return self.get_flag_ids_for_service(service_name)

    def __getitem__(self, service_name: str) -> ServiceScopedFlagIds:
        return self.get_flag_ids_for_service(service_name)



import difflib
import hashlib
import sys
from typing import Iterable

from loguru import logger

LOGGER_CONFIG = {
    "handlers": [
        {
            "sink": sys.stdout,
            "format": "<d>[{time:HH:mm:ss}]</d> <level>{level: <8}</level> {message}",
        },
    ],
}

COLORS = [
    # Red shades
    (255, 160, 122),
    (250, 128, 114),
    (240, 128, 128),
    (205, 92, 92),
    # Orange shades
    (255, 160, 122),
    (255, 127, 80),
    (255, 99, 71),
    (255, 69, 0),
    (255, 140, 0),
    (255, 165, 0),
    # Yellow shades
    (189, 183, 107),
    (240, 230, 140),
    (255, 218, 185),
    (250, 250, 210),
    (255, 215, 0),
    # Green shades
    (0, 128, 0),
    (46, 139, 87),
    (144, 238, 144),
    (50, 205, 50),
    (0, 255, 0),
    (173, 255, 47),
    # Blue shades
    (30, 144, 255),
    (135, 206, 235),
    (176, 224, 230),
    (70, 130, 180),
    (0, 206, 209),
    # Purple shades
    (219, 112, 147),
    (199, 21, 133),
    (255, 20, 147),
    (255, 105, 180),
    (255, 192, 203),
    # Pink shades
    (219, 112, 147),
    (199, 21, 133),
    (255, 20, 147),
    (255, 105, 180),
    (255, 192, 203),
    # Brown shades
    (210, 105, 30),
    (205, 133, 63),
    (188, 143, 143),
    (222, 184, 135),
    (255, 228, 196),
]

logger = logger.opt(colors=True)
logger.configure(**LOGGER_CONFIG)  # type: ignore

_loguru_error = logger.error


def _custom_error(msg: str, *args, **kwargs) -> None:
    """
    Wraps error message with <red> tags turning the message red.

    :param msg: Message to be logged
    :type msg: str
    """
    _loguru_error(f"<red>{msg}</>", *args, **kwargs)


logger.error = _custom_error  # type: ignore


def _hash_to_color(s: str) -> tuple[int, int, int]:
    """
    Returns a color tuple based on the hash of the input string.

    :param s: Input string
    :type s: str
    :return: Tuple of RGB values
    :rtype: tuple[int, int, int]
    """
    hash_hex = hashlib.md5(s.encode()).hexdigest()
    hash_dec = int(hash_hex, 16)
    return COLORS[hash_dec % len(COLORS)]


def colorize(text: str, reset: str = "white") -> str:
    """
    Colorizes a string deterministically based on its hash. The same string will always
    have the same color.

    :param text: Input text to colorize
    :type text: str
    :param reset: Name or sequence to reset text color, defaults to `white`. Use `red` for error messages and ANSI
    escape codes for custom colors.
    :type reset: str, optional
    :return: Colored text
    :rtype: str
    """
    r, g, b = _hash_to_color(text)

    reset_color_map = {
        "white": "\033[0m",
        "red": "\033[0;31m",
        "dim": "\033[0;2m",
    }

    colored_string = f"\033[38;2;{r};{g};{b}m{text}{reset_color_map.get(reset, reset)}"
    return colored_string


def suggest_closest_match(input_string: str, candidates: Iterable[str]) -> str:
    """
    Suggests the most similar string from an iterable of possible strings.

    :param input_string: Input string to find the closest match for.
    :type input_string: str
    :param candidates: Iterable of possible strings.
    :type candidates: Iterable[str]
    :return: Suggested string.
    :rtype: str
    """
    matches = difflib.get_close_matches(input_string, candidates, n=1)
    if matches:
        return matches[0]
    return ""


def truncate(input_string: str, length: int = 50) -> str:
    """
    Truncates a string to a specified length and appends an ellipsis if the string is longer.

    :param input_string: Input string to truncate.
    :type input_string: str
    :param length: Maximum length of the string.
    :type length: int
    :return: Truncated string.
    :rtype: str
    """
    return (input_string[:length] + "...") if len(input_string) > length else input_string



import base64
import pickle
from typing import Any

from ..api_client.schemas import FlagsEnqueueBody
from .base import HashRedisStorage, SetRedisStorage


class BlobStorage(HashRedisStorage[Any]):
    """
    Simple Redis key-value store for storing arbitrary objects.
    Objects are pickled before storing and unpickled when retrieved.
    """

    def _encode(self, value: Any) -> bytes:
        obj_blob = pickle.dumps(value)
        return base64.b64encode(obj_blob)

    def _decode(self, value: bytes) -> Any:
        obj_blob = base64.b64decode(value)
        return pickle.loads(obj_blob)


class FlagIdsHashStorage(SetRedisStorage[str]):
    """
    Simple Redis set store for storing obtained flag ID hashes.
    """

    def _encode(self, value: str) -> bytes:
        if not isinstance(value, str):
            raise TypeError("Value must be a string.")
        return value.encode("utf-8")

    def _decode(self, value: bytes) -> str:
        return value.decode("utf-8")


class UnsentFlagStorage(SetRedisStorage[FlagsEnqueueBody]):
    """
    Simple Redis set store for storing unsent flags.
    """

    def _encode(self, value: FlagsEnqueueBody) -> bytes:
        if not isinstance(value, FlagsEnqueueBody):
            raise TypeError("Value must be of type EnqueueBody.")
        return value.model_dump_json().encode("utf-8")

    def _decode(self, value: bytes) -> FlagsEnqueueBody:
        return FlagsEnqueueBody.model_validate_json(value.decode("utf-8"))
